@@ -325,27 +325,22 @@ pub enum PlotKind {
 #[derive(Debug, StructOpt, Serialize, Deserialize, Clone)]
 pub enum EstimateKind {
     #[structopt(
-        name = "tmb",
-        about = "Estimate tumor mutational burden. Takes Varlociraptor calls (must be annotated \
-                 with e.g. snpEFF) from STDIN, prints TMB estimate in Vega-lite JSON format to STDOUT. \
+        name = "mutational-burden",
+        about = "Estimate mutational burden. Takes Varlociraptor calls (must be annotated \
+                 with e.g. VEP but using ANN instead of CSQ) from STDIN, prints mutational burden estimate in Vega-lite JSON format to STDOUT. \
                  It can be converted to an image via vega-lite-cli (see conda package).",
-        usage = "varlociraptor estimate tmb --coding-genome-size 3e7 --somatic-tumor-events SOMATIC_TUMOR \
-                 --tumor-sample tumor < calls.bcf | vg2svg > tmb.svg",
+        usage = "varlociraptor estimate mutational-burden --coding-genome-size 3e7 --events SOMATIC_TUMOR \
+                 --sample tumor < calls.bcf | vg2svg > tmb.svg",
         setting = structopt::clap::AppSettings::ColoredHelp,
     )]
-    TMB {
+    MutationalBurden {
+        #[structopt(long = "events", help = "Events to consider (e.g. SOMATIC_TUMOR).")]
+        events: Vec<String>,
         #[structopt(
-            long = "somatic-tumor-events",
-            default_value = "SOMATIC_TUMOR",
-            help = "Events to consider (e.g. SOMATIC_TUMOR)."
+            long = "sample",
+            help = "Name(s) of the sample(s) in the given VCF/BCF. Multiple samples can be given when using the multibar plot mode."
         )]
-        somatic_tumor_events: Vec<String>,
-        #[structopt(
-            long = "tumor-sample",
-            default_value = "tumor",
-            help = "Name(s) of the tumor sample(s) in the given VCF/BCF. Multiple samples can be given when using the multibar plot mode."
-        )]
-        tumor_sample: Vec<String>,
+        sample: Vec<String>,
         #[structopt(
             long = "coding-genome-size",
             default_value = "3e7",
@@ -354,10 +349,10 @@ pub enum EstimateKind {
         coding_genome_size: f64,
         #[structopt(
             long = "plot-mode",
-            possible_values = &estimation::tumor_mutational_burden::PlotMode::iter().map(|v| v.into()).collect_vec(),
+            possible_values = &estimation::mutational_burden::PlotMode::iter().map(|v| v.into()).collect_vec(),
             help = "How to plot (as stratified curve, histogram or multi-sample barplot)."
         )]
-        mode: estimation::tumor_mutational_burden::PlotMode,
+        mode: estimation::mutational_burden::PlotMode,
         #[structopt(
             long = "vaf-cutoff",
             default_value = "0.2",
@@ -999,15 +994,15 @@ pub fn run(opt: Varlociraptor) -> Result<()> {
             conversion::decode_phred::decode_phred()?;
         }
         Varlociraptor::Estimate { kind } => match kind {
-            EstimateKind::TMB {
-                somatic_tumor_events,
-                tumor_sample,
+            EstimateKind::MutationalBurden {
+                events,
+                sample,
                 coding_genome_size,
                 mode,
                 cutoff,
-            } => estimation::tumor_mutational_burden::collect_estimates(
-                &somatic_tumor_events,
-                &tumor_sample,
+            } => estimation::mutational_burden::collect_estimates(
+                &events,
+                &sample,
                 coding_genome_size as u64,
                 mode,
                 cutoff as f64,
